--[[
    File: TuckUnitFramesAuraFrame.lua
    Author: Tuck
    Description: This file contains all funtions that manage aura (buff or debuff) frames.

]]


--[[
    Libraries
]]
local IFrameFactory = IFrameFactory("1.0")
local Dewdrop = AceLibrary("Dewdrop-2.0")
local TRACE = function(msg) ChatFrame1:AddMessage(msg) end


-- namespace
local tuf_auraframe = TuckUF.auraframe
local tuf_unitframe = TuckUF.unitframe
local tuf_network = TuckUF.network
local tuf_aura_db = TuckUF.aura_db

--[[
    Enables/disables the buff frame overlay. The overlay can be clicked for options.
    Deprecated.
]]

function TuckUF.auraframe:SetOverlayEnabled(frame, enabled)
    if enabled then
        frame:Show()
    else
        frame:Hide()
    end
end

--[[
    Toggles all overlays
    Deprecated.
]]
function TuckUF.auraframe:SetOverlaysEnabled(enabled)
    for k, v in pairs(TuckUF.auraframes) do
        tuf_auraframe:SetOverlayEnabled(v, enabled)
    end
end

--[[
    Returns a table that holds the options for this frame.
    Options that are not present are set to their default values
]]
function TuckUF.auraframe:GetFrameOption(frame, option)
    local db = TuckUF:GetDB()
    return db[frame.config.."_"..option]   
end

--[[
	Reads the anchoring information for the given frame
	and decides where to place the first icon.
]]
local function AuraFrame_SetLeadAnchor(frame)
	frame:ClearAllPoints()       
           
    -- attach the frame to the parent frame
    if frame.anchor == "TOPLEFT" then
        frame:SetPoint("BOTTOMLEFT", frame:GetParent(), "TOPLEFT", 0, 0)
        frame.first_corner = "BOTTOMRIGHT"
    elseif frame.anchor == "TOPRIGHT" then
        frame:SetPoint("BOTTOMRIGHT", frame:GetParent(), "TOPRIGHT", 0, 0)
        frame.first_corner = "BOTTOMRIGHT"
    elseif frame.anchor == "RIGHTTOP" then
        frame:SetPoint("TOPLEFT", frame:GetParent(), "TOPRIGHT", 0, 0)
        frame.first_corner = "TOPLEFT"
    elseif frame.anchor == "RIGHTBOTTOM" then
        frame:SetPoint("BOTTOMLEFT", frame:GetParent(), "BOTTOMRIGHT", 0, 0)
        frame.first_corner = "BOTTOMLEFT"
    elseif frame.anchor == "BOTTOMRIGHT" then
        frame:SetPoint("TOPRIGHT", frame:GetParent(), "BOTTOMRIGHT", 0, 0)
        frame.first_corner = "TOPRIGHT"
    elseif frame.anchor == "BOTTOMLEFT" then
        frame:SetPoint("TOPLEFT", frame:GetParent(), "BOTTOMLEFT", 0, 0)
        frame.first_corner = "TOPLEFT"
    elseif frame.anchor == "LEFTBOTTOM" then
        frame:SetPoint("BOTTOMRIGHT", frame:GetParent(), "BOTTOMLEFT", 0, 0)
        frame.first_corner = "BOTTOMRIGHT"
    elseif frame.anchor == "LEFTTOP" then
        frame:SetPoint("TOPRIGHT", frame:GetParent(), "TOPLEFT", 0, 0)
        frame.first_corner = "TOPRIGHT"
    elseif frame.anchor == "INSIDELEFT" then
        frame:SetPoint("LEFT", frame:GetParent(), "LEFT", 0, 0)
        frame.first_corner = "TOPLEFT"
    elseif frame.anchor == "INSIDERIGHT" then
        frame:SetPoint("RIGHT", frame:GetParent(), "RIGHT", 0, 0)
        frame.first_corner = "TOPRIGHT"     
    elseif frame.anchor == "LEFT" then
        frame:SetPoint("RIGHT", frame:GetParent(), "LEFT", 0, 0)
        frame.first_corner = "TOPRIGHT"     
    elseif frame.anchor == "RIGHT" then
        frame:SetPoint("LEFT", frame:GetParent(), "RIGHT", 0, 0)
        frame.first_corner = "TOPLEFT"             
    else
        assert(false)
    end
end	


--[[
	Sets the anchoring information for the icons. This
	depends on the anchoring of the lead icon.
	Example: if the lead icon is located in the upper left corner, 
	the icons are filled left-to-right and then top-to-bottom.
	Always fills horizontally first.
]]
local function AuraFrame_SetIconAnchor(frame)   
	if frame.first_corner == "TOPLEFT" then
		frame.lead_anchor = "TOPLEFT"
		frame.lead_relative_to = "TOPLEFT"
		frame.icon_anchor = "LEFT"
		frame.icon_relative_to = "RIGHT"
		frame.row_anchor = "TOPLEFT"
		frame.row_relative_to = "BOTTOMLEFT"
		frame.icon_dx = frame.icon_gap
		frame.icon_dy = -frame.icon_gap
	elseif frame.first_corner == "TOPRIGHT" then
		frame.lead_anchor = "TOPRIGHT"
		frame.lead_relative_to = "TOPRIGHT"
		frame.icon_anchor = "RIGHT"
		frame.icon_relative_to = "LEFT"
		frame.row_anchor = "TOPRIGHT"
		frame.row_relative_to = "BOTTOMRIGHT"
		frame.icon_dx = -frame.icon_gap
		frame.icon_dy = -frame.icon_gap
	elseif frame.first_corner == "BOTTOMLEFT" then
		frame.lead_anchor = "BOTTOMLEFT"
		frame.lead_relative_to = "BOTTOMLEFT"
		frame.icon_anchor = "LEFT"
		frame.icon_relative_to = "RIGHT"
		frame.row_anchor = "BOTTOMLEFT"
		frame.row_relative_to = "TOPLEFT"
		frame.icon_dx = frame.icon_gap
		frame.icon_dy = frame.icon_gap
	elseif frame.first_corner == "BOTTOMRIGHT" then
		frame.lead_anchor = "BOTTOMRIGHT"
		frame.lead_relative_to = "BOTTOMRIGHT"
		frame.icon_anchor = "RIGHT"
		frame.icon_relative_to = "LEFT"
		frame.row_anchor = "BOTTOMRIGHT"
		frame.row_relative_to = "TOPRIGHT"
		frame.icon_dx = -frame.icon_gap
		frame.icon_dy = frame.icon_gap
	else
		-- exception
        assert(false);
	end
end


--[[
	Destroys all icon frames for the given frame.
]]
local function AuraFrame_DestroyIcons(frame)
	if (frame.icons) then
		for _, iconFrame in pairs(frame.icons) do
			IFrameFactory:Destroy("TuckUF", "AuraIcon", iconFrame)
		end
	end
end

--[[
	Reads the icon size information
]]
local function AuraFrame_ReadOptions(frame)
    frame.enabled = tuf_auraframe:GetFrameOption(frame, "enabled") or false
    frame.row_count = tuf_auraframe:GetFrameOption(frame, "rows") or 1
    frame.column_count = tuf_auraframe:GetFrameOption(frame, "columns") or 1
    frame.icon_size = tuf_auraframe:GetFrameOption(frame, "icon_size") or 16
    frame.icon_gap = tuf_auraframe:GetFrameOption(frame, "icon_gap") or 0
    frame.anchor = tuf_auraframe:GetFrameOption(frame, "anchor") or "BOTTOMLEFT"
    frame.transparency = tuf_auraframe:GetFrameOption(frame, "transparency") or 0
    frame.offset_x = tuf_auraframe:GetFrameOption(frame, "offset_x") or 0
    frame.offset_y = tuf_auraframe:GetFrameOption(frame, "offset_y") or 0
    
    frame.filters = {}
    frame.filters.source_foreign = tuf_auraframe:GetFrameOption(frame, "filter_foreign")
    frame.filters.source_own = tuf_auraframe:GetFrameOption(frame, "filter_own")
    frame.filters.cureable_yes = tuf_auraframe:GetFrameOption(frame, "filter_cureable")
    frame.filters.cureable_no = tuf_auraframe:GetFrameOption(frame, "filter_noncureable")
    frame.filters.power_small = tuf_auraframe:GetFrameOption(frame, "filter_weak")
    frame.filters.power_medium = tuf_auraframe:GetFrameOption(frame, "filter_normal")
    frame.filters.power_high = tuf_auraframe:GetFrameOption(frame, "filter_powerful")
    frame.filters.group_yes = tuf_auraframe:GetFrameOption(frame, "filter_group")
    frame.filters.group_no = tuf_auraframe:GetFrameOption(frame, "filter_nogroup")
    
    frame.group_mode = TuckUF.auraframe:GetFrameOption(frame, "group_mode")
    frame.sort_mode = TuckUF.auraframe:GetFrameOption(frame, "sort_mode")
    
    frame.icon_count = frame.row_count * frame.column_count
    
    frame:SetWidth(frame.column_count * frame.icon_size + (frame.column_count+1) * frame.icon_gap)
    frame:SetHeight(frame.row_count * frame.icon_size + (frame.row_count+1) * frame.icon_gap)
end

--[[
	Creates all icons.
]]
local function AuraFrame_CreateIcons(frame)  
	frame.icons = {}
	
	local i
	for i = 1, frame.icon_count do
        -- create frame
		local iconFrame = IFrameFactory:Create("TuckUF", "AuraIcon")
		frame.icons[i] = iconFrame
        iconFrame:SetParent(frame:GetParent())
        iconFrame:SetFrameLevel(frame:GetFrameLevel() + 1)
		iconFrame:SetWidth(frame.icon_size)
		iconFrame:SetHeight(frame.icon_size)
        iconFrame:SetAlpha((100.0 - frame.transparency)/100.0);
        iconFrame.duration:SetWidth(iconFrame:GetWidth())
		
		-- member variables
		iconFrame.id = i
		iconFrame.aura_id = 0
		iconFrame.unit = frame.unit
		
		-- tooltips
		if frame.shows_buffs then
			iconFrame:SetScript("OnEnter",function() GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT"); GameTooltip:SetUnitBuff(this.unit, this.aura_id)  end)
		else
			iconFrame:SetScript("OnEnter",function() GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT"); GameTooltip:SetUnitDebuff(this.unit, this.aura_id)  end)
		end
		iconFrame:SetScript("OnLeave",function() GameTooltip:Hide() end)
        iconFrame:EnableMouse(true)
		
		-- anchoring in columns and rows
        iconFrame:ClearAllPoints()
		if i == 1 then
			iconFrame:SetPoint(frame.lead_anchor, frame, frame.lead_relative_to, frame.offset_x, frame.offset_y)
		elseif i % frame.column_count == 1 or frame.column_count == 1 then
			iconFrame:SetPoint(frame.row_anchor, frame.icons[i - frame.column_count], frame.row_relative_to, 0, frame.icon_dy)
		else
			iconFrame:SetPoint(frame.icon_anchor, frame.icons[i - 1], frame.icon_relative_to, frame.icon_dx, 0)
		end
		
		iconFrame:Show()
	end
end

--[[
    Call after the frame options have changed
]]

function TuckUF.auraframe:OnInit(frame)
    -- destroy old icons
	AuraFrame_DestroyIcons(frame)
	
    -- read options
	AuraFrame_ReadOptions(frame)
    
    -- set anchors
    AuraFrame_SetLeadAnchor(frame)
	AuraFrame_SetIconAnchor(frame)
    
    -- create new icons
    AuraFrame_CreateIcons(frame)
    
    -- update contents
    tuf_auraframe:UpdateAuras(frame)
end

--[[
	Returns the i-th aura (buff or debuff, depending on frame modus)
]]
local function get_frame_aura(frame, i)
	if frame.shows_buffs then
		local name, rank, texture, applications, duration, timeleft = UnitBuff(frame.unit, i)
		return name, rank, texture, applications, nil, duration, timeleft
	else
		local name, rank, texture, applications, type, duration, timeleft = UnitDebuff(frame.unit, i)
		return name, rank, texture, applications, type, duration, timeleft
	end
end

--[[
    insertion sort: move item at the given index towards the head
    pre: [0, index - 1] is sorted. post: [0, index] is sorted
]]
local function list_sort_item(list, index, comp)
	local i = index
	while i > 0 and comp(list[i]) > comp(list[i - 1]) do
		list_swap(list, i, i - 1)
		i = i - 1
	end
end

--[[
    sorts the list
]]
local function list_sort(list, comp)
	local i
	for i = 0, list.size do
		list_sort_item(list, i, comp)
	end
end

--[[
	Processes the list, grouping its items. All items that are classified as equal according to the
	given function are replaced by a single item.
]]
local function list_group_items(list, equalFunc, assignFunc)
	local i = 1
	while i <= #list do
        local found = false
        local list_i = list[i]
		for j = 1, i-1 do
            local list_j = list[j]
			if equalFunc(list_i, list_j) then
                list_j.applications = list_j.applications + list_i.applications
				list[j] = assignFunc(list_i, list_j)
                found = true
				table.remove(list, i)
				break
			end
		end
        if not found then
            i = i + 1
        end
	end
end

--[[ 
    Local buff table
]]
local buff_table = {}

for i = 1, 40 do
	buff_table[i] = {}
end

--[[ 
    Buff equality.
]]

local buff_equal_functions = {}

buff_equal_functions["name"] = function (buff1, buff2)
	return buff1.name == buff2.name
end

buff_equal_functions["group"] = function (buff1, buff2)
	return buff1.group_name == buff2.group_name and buff1.group_name ~= ""
end


--[[
	Buff ordering.
]]

local buff_order_functions = {}

-- greatest power first
buff_order_functions["max_power"] = function (buff1, buff2)
	return buff1.power > buff2.power
end

-- smallest power first
buff_order_functions["min_power"] =  function (buff1, buff2)
	return buff1.power < buff2.power
end

-- shortest duration first
buff_order_functions["min_duration"] =  function (buff1, buff2)
	return 
        buff1.duration and buff2.duration and buff1.duration < buff2.duration
        or
        buff1.duration ~= nil and buff2.duration == nil
end

-- longest duration first
buff_order_functions["max_duration"] =  function (buff1, buff2)
	return 
        buff1.duration and buff2.duration and buff1.duration > buff2.duration
        or
        buff1.duration == nil and buff2.duration ~= nil
end

-- shorter time left first
buff_order_functions["min_timeleft"] =  function (buff1, buff2)
	return 
        buff1.timeleft and buff2.timeleft and buff1.timeleft < buff2.timeleft
        or
        buff1.timeleft ~= nil and buff2.timeleft == nil
end


-- longer time left first
buff_order_functions["max_timeleft"] =  function (buff1, buff2)
	return 
        buff1.timeleft and buff2.timeleft and buff1.timeleft > buff2.timeleft
        or
        buff1.timeleft == nil and buff2.timeleft ~= nil
end

--[[
    Buff filtering
]]
local buff_filter_functions = {}


buff_filter_functions["source_own"] = function(buff)
    return buff.is_own
end

buff_filter_functions["source_foreign"] = function(buff)
    return not buff.is_own
end

buff_filter_functions["cureable_yes"] = function(buff)
    return buff.is_cureable
end

buff_filter_functions["cureable_no"] = function(buff)
    return not buff.is_cureable
end

buff_filter_functions["power_high"] = function(buff)
    return buff.power >= 200
end

buff_filter_functions["power_medium"] = function(buff)
    return buff.power >= 100 and buff.power < 200
end

buff_filter_functions["power_small"] = function(buff)
    return buff.power < 100
end

buff_filter_functions["group_yes"] = function(buff)
    return buff.group_name ~= nil and buff.group_name ~= "" and buff.group_name ~= "none"
end

buff_filter_functions["group_no"] = function(buff)
    return buff.group_name == nil or buff.group_name == "" or buff.group_name == "none"
end

--[[
	Call after creating the frame
]]
function TuckUF.auraframe:CreateFrame(unitframe, modus)
	if not (modus == "buffs" or modus == "debuffs" or modus == "buffs2"
		or modus == "debuffs2") then
		
		return nil
	end
	
    local unit = unitframe.unit
    local config = unitframe.config
    
	local frame = CreateFrame("Button", unitframe:GetName().."_AuraFrame_"..modus, unitframe)
    
	frame:SetBackdrop({
	bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true, tileSize = 12, edgeSize = 12,
	insets = { left = 2, right = 2, top = 2, bottom = 2 }})
    
    frame:SetFrameLevel(unitframe:GetFrameLevel()+20)
	
	frame:SetBackdropBorderColor(0.8, 0.8, 0.8, 1)
	frame:SetBackdropColor(0, 0, 0, 1)
	
    frame.config = config.."_"..modus
	frame.unit = unit
	frame.modus = modus
	frame.show_uninteresting = true
	frame.show_foreign = true
	frame.show_uncureable = true
	frame.shows_buffs = (modus == "buffs" or modus == "buffs2")
	frame.shows_debuffs = (modus == "debuffs" or modus == "debuffs2")
	
	frame:SetWidth(100)
    frame:SetHeight(100)
    frame:SetPoint("CENTER", UIParent)
	   
	frame:Hide()
	
	return frame
end


--[[
	Updates all auras for the given frame
]]
function TuckUF.auraframe:UpdateAuras(frame)
    -- check whether all auras should be hidden
    if not frame.enabled then
        for i = 1, frame.icon_count do
            frame.icons[i]:Hide()
        end
        return
    end
    
	-- true if unit is friend
	local is_friend = UnitIsFriend("player", frame.unit)
    local aura_info_suffix = tuf_aura_db:GetOptionSuffix(frame.unit)
	
	-- clear buff list
	local filtered_buffs = {}
    
    -- clear data
    frame.has_special_aura = false
    frame.special_aura_power = -1
	
	-- fill buff list with buffs
	for i = 1, 40 do
		-- get buff object reference
		local buff = buff_table[i]
		
		-- get buff info from WoW API
		buff.name, buff.rank, buff.texture, buff.applications, buff.type, buff.duration, buff.timeleft = get_frame_aura(frame, i)

		if buff.name then
					
			-- compute additional buff info
            local buff_info = tuf_aura_db:GetAuraInfo(buff.name)
			buff.id = i
			buff.is_own = not (buff.duration == nil)
			buff.is_cureable = buff.type and TuckUF.DispelTypes[buff.type] or false
			buff.power = buff_info["power"..aura_info_suffix] or 0
			buff.group_name = buff_info["group"..aura_info_suffix] or ""
            buff.is_special = buff_info["is_special"..aura_info_suffix] or false
           
            -- remember whether this frame has a special/cureable aura
            frame.has_special_aura = frame.has_special_aura or buff.is_special or buff.is_cureable
            
            -- set the border color to the respective color
            if buff.power > frame.special_aura_power then
                if buff.is_special then
                    frame.special_aura_power = buff.power
                    frame.special_aura_color = TuckUF.DebuffColors["Special"]
                elseif buff.is_cureable then
                    frame.special_aura_power = buff.power
                    frame.special_aura_color = TuckUF.DebuffColors[buff.type]                
                end
            end
			
			-- try to get timing from other players if missing
			if not buff.duration and not buff.timeleft then
				buff.timeleft, buff.duration = tuf_network:GetAuraTimeByUnitID(frame.unit, buff.name)
			end
		
			-- true if this buff/debuff is to be shown                    
            local show_aura = true
            for filter_name, filter_enabled in pairs(frame.filters) do
                if filter_enabled and buff_filter_functions[filter_name](buff) then
                    show_aura = false
                    break
                end
            end
					
			-- if yes, add to list
			if (show_aura) then
				table.insert(filtered_buffs, buff)
			end
        else
            -- no more auras present
            break
		end
	end
	
	-- group buff list
    local equal_function = buff_equal_functions[frame.group_mode]
    local assign_function = nil
    if group_function then
        list_group_items(filtered_buffs, equal_function, buff_assign_power)
    end
	
	-- sort buff list
    local order_function = buff_order_functions[frame.sort_mode]
    if order_function then
        table.sort(filtered_buffs, order_function)
    end
    
	-- use buff list to fill icons
	for i, buff in ipairs(filtered_buffs) do
	
		-- icon reference
		local auraFrame = frame.icons[i]
        
        -- if icon does not exist, we have run out of icons
        if auraFrame == nil then 
            break
        end
				
		-- buff id, for tooltip
		auraFrame.aura_id = buff.id
		
		-- texture	
        auraFrame.icon:SetTexture(buff.texture)
			
		-- display stacks
		if buff.applications > 1 then
			auraFrame.count:SetText(buff.applications)
		else
			auraFrame.count:SetText("")
		end
		
		-- display duration
		auraFrame.aura_duration = buff.duration
		auraFrame.aura_timeleft = buff.timeleft
		auraFrame.aura_timestamp = GetTime()
		auraFrame:onUpdate()
		
		-- show
		auraFrame:Show()
	end
	  
	for i = #filtered_buffs + 1, frame.icon_count do
		frame.icons[i]:Hide()
	end
end
